'''
Created on Sep 1, 2017

@author: sumanth-3058
'''
try:
    from .RestClient import ZCRMRestClient
    from .Operations import ZCRMAttachment,ZCRMCustomView,ZCRMCustomViewCategory,ZCRMCustomViewCriteria,ZCRMEventParticipant,ZCRMField,ZCRMInventoryLineItem,ZCRMJunctionRecord,ZCRMLayout,ZCRMLeadConvertMapping,ZCRMLeadConvertMappingField,ZCRMLookupField,ZCRMModule,ZCRMModuleRelatedList,ZCRMModuleRelation,ZCRMNote,ZCRMPermission,ZCRMPickListValue,ZCRMPriceBookPricing,ZCRMProfile,ZCRMProfileCategory,ZCRMProfileSection,ZCRMRecord,ZCRMRelatedListProperties,ZCRMRole,ZCRMSection,ZCRMTax,ZCRMTrashRecord,ZCRMUser,ZCRMUserCustomizeInfo,ZCRMUserTheme,ZCRMVariable,ZCRMVariableGroup,ZCRMBlueprint,ZCRMBlueprintRelatedModule,ZCRMBlueprintTransition,ZCRMTransitionAttachment,ZCRMTransitionCheckList,ZCRMTransitionRelatedData,ZCRMTransition,ZCRMTransitionCrypt,ZCRMOrgTax,ZCRMTag
    from .Org import ZCRMOrganization
    from .CLException import ZCRMException
    from .Utility import APIConstants
    from .OAuthClient import ZohoOAuth,ZohoOAuthClient,ZohoOAuthTokens,AbstractZohoOAuthPersistence
except ImportError:
    from RestClient import ZCRMRestClient
    from Operations import ZCRMAttachment,ZCRMCustomView,ZCRMCustomViewCategory,ZCRMCustomViewCriteria,ZCRMEventParticipant,ZCRMField,ZCRMInventoryLineItem,ZCRMJunctionRecord,ZCRMLayout,ZCRMLeadConvertMapping,ZCRMLeadConvertMappingField,ZCRMLookupField,ZCRMModule,ZCRMModuleRelatedList,ZCRMModuleRelation,ZCRMNote,ZCRMPermission,ZCRMPickListValue,ZCRMPriceBookPricing,ZCRMProfile,ZCRMProfileCategory,ZCRMProfileSection,ZCRMRecord,ZCRMRelatedListProperties,ZCRMRole,ZCRMSection,ZCRMTax,ZCRMTrashRecord,ZCRMUser,ZCRMUserCustomizeInfo,ZCRMUserTheme,ZCRMVariable,ZCRMVariableGroup,ZCRMBlueprint,ZCRMBlueprintRelatedModule,ZCRMBlueprintTransition,ZCRMTransitionAttachment,ZCRMTransitionCheckList,ZCRMTransitionRelatedData,ZCRMTransition,ZCRMTransitionCrypt,ZCRMOrgTax,ZCRMTag
    from Org import ZCRMOrganization
    from CLException import ZCRMException
    from Utility import APIConstants
    from OAuthClient import ZohoOAuth,ZohoOAuthClient,ZohoOAuthTokens,AbstractZohoOAuthPersistence


'''
Created on Aug 16, 2017

@author: sumanth-3058
'''

class ZohoOAuthConstants(object):
    '''
    OAuth constants
    '''
    IAM_URL="accounts_url";
    SCOPES="scope";
    STATE="state";
    STATE_OBTAINING_GRANT_TOKEN="OBTAIN_GRANT_TOKEN";
    RESPONSE_TYPE="response_type";
    RESPONSE_TYPE_CODE="code";
    CLIENT_ID="client_id";
    CLIENT_SECRET="client_secret";
    REDIRECT_URL="redirect_uri";
    ACCESS_TYPE="access_type";
    ACCESS_TYPE_OFFLINE="offline";
    ACCESS_TYPE_ONLINE="online";
    PROMPT="prompt";
    PROMPT_CONSENT="consent";
    GRANT_TYPE="grant_type";
    GRANT_TYPE_AUTH_CODE="authorization_code";
    TOKEN_PERSISTENCE_PATH="token_persistence_path";
    CUSTOM_PERSISTENCE_HANDLER_PATH = "persistence_handler_path";
    CUSTOM_PERSISTENCE_HANDLER_CLASS = "persistence_handler_class";
    SANDBOX="sandbox";
    DATABASE_PORT="mysql_port";
    DATABASE_PASSWORD="mysql_password";
    DATABASE_USERNAME="mysql_username";
    PERSISTENCE_FILE_NAME="zcrm_oauthtokens.pkl"
    
    GRANT_TYPE_REFRESH="refresh_token";
    CODE="code";
    GRANT_TOKEN="grant_token";
    ACCESS_TOKEN="access_token";
    REFRESH_TOKEN="refresh_token";
    EXPIRES_IN = "expires_in";
    EXPIRIY_TIME = "expiry_time";
    TOKEN = "token";
    DISPATCH_TO = "dispatchTo";
    OAUTH_TOKENS_PARAM = "oauth_tokens";
    
    OAUTH_HEADER_PREFIX="Zoho-oauthtoken ";
    AUTHORIZATION="Authorization";
    REQUEST_METHOD_GET="GET";
    REQUEST_METHOD_POST="POST";
    
    RESPONSECODE_OK=200;

class ZohoOAuthException(Exception):
    '''
    This is the custom exception class for handling Client Library OAuth related exceptions 
    '''
    def __init__(self, err_message):
        self.message=err_message
        Exception.__init__(self,err_message)

    def __str__(self):
        return self.message

class ZohoOAuthHTTPConnector(object):
    '''
    This module is to make HTTP connections, trigger the requests and receive the response
    '''
    @staticmethod
    def get_instance(url,params=None,headers=None,body=None,method=None):
        return ZohoOAuthHTTPConnector(url,params,headers,body,method)
    
    def __init__(self, url,params,headers,body,method):
        '''
        Constructor
        '''
        self.url=url
        self.req_headers=headers
        self.req_method=method
        self.req_params=params
        self.req_body=body
        
    def trigger_request(self):
        response=None
        import requests,json
        if(self.req_method == ZohoOAuthConstants.REQUEST_METHOD_GET):
            response=requests.get(self.url,params=self.req_params, headers=self.req_headers,allow_redirects=False)
        elif(self.req_method==ZohoOAuthConstants.REQUEST_METHOD_POST):
            response=requests.post(self.url,data=json.dumps(self.req_body), params=self.req_params,headers=self.req_headers,allow_redirects=False)
        return response
    def set_url(self,url):
        self.url=url
    def get_url(self):
        return self.url
    def add_http_header(self,key,value):
        self.req_headers[key]=value
    def get_http_headers(self):
        return self.req_headers
    def set_http_request_method(self,method):
        self.req_method=method
    def get_http_request_method(self):
        return self.req_method
    def set_request_body(self,reqBody):
        self.req_body=reqBody
    def get_request_body(self):
        return self.req_body
    def add_http_request_params(self,key,value):
        self.req_params[key]=value
    def get_http_request_params(self):
        return self.req_params
    
class ZohoOAuthParams(object):
    '''
    This class is to OAuth related params(i.e. client_id,client_secret,..)
    '''
    def __init__(self, client_id,client_secret,redirect_uri):
        '''
        Constructor
        '''
        self.clientID=client_id
        self.clientSecret=client_secret
        self.redirectUri=redirect_uri
    @staticmethod
    def get_instance(client_id,client_secret,redirect_uri):
        return ZohoOAuthParams(client_id,client_secret,redirect_uri)
    
import logging
logger=logging.getLogger('Client_Library_OAUTH')
class OAuthLogger(object):
    '''
    This class is to log the exceptions onto console and file
    '''
    @staticmethod
    def add_log(message,level,exception=None):
        logger.setLevel(logging.DEBUG)
        consoleHandler = logging.StreamHandler()
        # create formatter and add it to the handlers
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        
        try:
            from .Utility import ZCRMConfigUtil,APIConstants
        except ImportError:
            from Utility import ZCRMConfigUtil,APIConstants
        log_path = None
        if(APIConstants.APPLICATION_LOGFILE_PATH in ZCRMConfigUtil.config_prop_dict):
            log_path=ZCRMConfigUtil.config_prop_dict[APIConstants.APPLICATION_LOGFILE_PATH]
        if log_path is not None and log_path.strip()!="":
            import os
            log_path=os.path.join(log_path,'oauth.log')
        
            fileHandler=logging.FileHandler(log_path)
            fileHandler.setLevel(logging.DEBUG)
            fileHandler.setFormatter(formatter)
            logger.addHandler(fileHandler)
        
        consoleHandler.setLevel(logging.DEBUG)
        consoleHandler.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(consoleHandler)
        
        
        if(exception!=None):
            message+='; Exception Message::'+exception.__str__()
        if(level==logging.ERROR):
            logger.error(message)
        elif(level==logging.INFO):
            logger.info(message)
        elif(level==logging.WARNING):
            logger.warning(message)


'''
Created on Aug 16, 2017

@author: sumanth-3058
'''
class ZCRMModule(object):
    '''
    This class is to deal with Zoho CRM modules
    '''
    def __init__(self, module_apiname):
        '''
        Constructor
        '''
        self.api_name=module_apiname
        self.is_convertable=None
        self.is_creatable=None
        self.is_editable=None
        self.is_deletable=None
        self.web_link=None
        self.singular_label=None
        self.plural_label=None
        self.modified_by=None
        self.modified_time=None
        self.is_viewable=None
        self.is_api_supported=None
        self.is_custom_module=None
        self.is_scoring_supported=None
        self.id=None
        self.module_name=None
        self.business_card_field_limit=None
        self.business_card_fields=list()
        self.profiles=list()
        self.display_field_name=None
        self.display_field_id=None
        self.related_lists=None
        self.layouts=None
        self.fields=None
        self.related_list_properties=None
        self.properties=None
        self.per_page=None
        self.search_layout_fields=None
        self.default_territory_name=None
        self.default_territory_id=None
        self.default_custom_view_id=None
        self.default_custom_view=None
        self.is_global_search_supported=None
        self.sequence_number=None
    
    @staticmethod
    def get_instance(module_apiname):
        return ZCRMModule(module_apiname)
    def get_record(self,entityID):
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        record=ZCRMRecord.get_instance(self.api_name, entityID)
        return EntityAPIHandler.get_instance(record).get_record()
    def get_records(self,cvid=None,sort_by=None,sort_order=None,page=0,per_page=200,custom_headers=None, custom_parameters=None):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).get_records(cvid,sort_by,sort_order,page,per_page,custom_headers, custom_parameters)
    def create_records(self,record_ins_list):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).create_records(record_ins_list)
    def upsert_records(self,record_ins_list,duplicate_check_fields=None):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).upsert_records(record_ins_list,duplicate_check_fields)
    def update_records(self,record_ins_list):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).update_records(record_ins_list)
    def mass_update_records(self,entityid_list,field_api_name,value):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).update_mass_records(entityid_list,field_api_name,value)
    def delete_records(self,entityid_list):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).delete_records(entityid_list)
    def get_all_deleted_records(self, page=0, per_page=200, custom_headers=None):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).get_all_deleted_records(page, per_page, custom_headers)
    def get_recyclebin_records(self, page=0, per_page=200, custom_headers=None):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).get_recyclebin_records(page, per_page, custom_headers)
    def get_permanently_deleted_records(self, page=0, per_page=200, custom_headers=None):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).get_permanently_deleted_records(page, per_page, custom_headers)
    def search_records(self,search_word,page=0,per_page=200):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).search_records(search_word,page,per_page,'word')
    def search_records_by_phone(self,phone,page=0,per_page=200):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).search_records(phone,page,per_page,'phone')
    def search_records_by_email(self,email,page=0,per_page=200):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).search_records(email,page,per_page,'email')
    def search_records_by_criteria(self,criteria,page=0,per_page=200):
        try:
            from .Handler import MassEntityAPIHandler
        except ImportError:
            from Handler import MassEntityAPIHandler
        return MassEntityAPIHandler.get_instance(self).search_records(criteria,page,per_page,'criteria')
    def get_all_fields(self):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).get_all_fields()
    def get_field(self,field_id):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).get_field(field_id)
    def get_all_layouts(self):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).get_all_layouts()
    def get_layout(self,layout_id):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).get_layout(layout_id)
    def get_all_customviews(self):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).get_all_customviews()
    def get_customview(self,customview_id):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).get_customview(customview_id)
    def get_all_relatedlists(self):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).get_all_relatedlists()
    def get_relatedlist(self,relatedlist_id):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).get_relatedlist(relatedlist_id)
    def update_module_settings(self):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).update_module_settings()
    def update_customview(self,customview_instance):
        try:
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Handler import ModuleAPIHandler
        return ModuleAPIHandler.get_instance(self).update_customview(customview_instance)
    def get_tags(self):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance(self).get_tags()

    def get_tag_count(self,tag_id):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance(self).get_tag_count(tag_id)

    def create_tags(self,tag_instances):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance(self).create_tags(tag_instances)

    def update_tags(self,tag_instances):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance(self).update_tags(tag_instances)

    def add_tags_to_multiple_records(self,tags_name_list, record_id_list):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance(self).add_tags_to_multiple_records(tags_name_list, record_id_list)

    def remove_tags_from_multiple_records(self,tags_name_list, record_id_list):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance(self).remove_tags_from_multiple_records(tags_name_list, record_id_list)

class ZCRMRecord(object):
    '''
    This class is to deal with Zoho CRM entity records
    '''
    def __init__(self,module_apiname,entity_id):
        '''
        Constructor
        '''
        self.module_api_name=module_apiname
        self.entity_id=entity_id
        self.line_items=[]
        self.lookup_label=None
        self.owner=None
        self.created_by=None
        self.modified_by=None
        self.created_time=None
        self.modified_time=None
        self.field_data=dict()
        self.properties = dict()
        self.participants = []
        self.price_details = []
        self.layout=None
        self.tax_list=[]
        self.tag_list=[]
        self.last_activity_time=None
        self.blueprint_transition_id = None
        self.blueprint_checklist = list()
        self.blueprint_values = dict()

    def set_checklist_value(self, key, value):
        d = dict()
        d[key] = value
        self.blueprint_checklist.append(d)

    def set_blueprint_data(self, key, value):
        self.blueprint_values[key] = value

    def get_checklist_values(self):
        return self.blueprint_checklist

    def get_blueprint_data(self):
        return self.blueprint_values

    @staticmethod
    def get_instance(module_apiname,entity_id=None):
        return ZCRMRecord(module_apiname,entity_id)
    def set_field_value(self,key,value):
        self.field_data[key]=value
    def get_field_value(self,key):
        return self.field_data[key] if key in self.field_data else None
    def add_line_item(self,line_item):
        self.line_items.append(line_item)
    
    def get(self):
        if self.entity_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Record_Get',"id should be set for the record",'ID DOES NOT PROVIDED',"ID")
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        return EntityAPIHandler.get_instance(self).get_record()
    def create(self):
        if self.entity_id is not None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Record_Create',"id should not be set for the record",'ID PROVIDED',"ID")
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        return EntityAPIHandler.get_instance(self).create_record()
    def update(self):
        if self.entity_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Record_Update',"id should be set for the record",'ID DOES NOT PROVIDED',"ID")
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        return EntityAPIHandler.get_instance(self).update_record()
    def delete(self):
        if self.entity_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Record_Delete',"id should be set for the record",'ID DOES NOT PROVIDED',"ID")
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        return EntityAPIHandler.get_instance(self).delete_record()
    def convert(self,potential_record=None, assign_to_user=None):
        if self.entity_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Convert_Record',"id should be set for the record",'ID DOES NOT PROVIDED',"ID")
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        return EntityAPIHandler.get_instance(self).convert_record(potential_record, assign_to_user)

    def get_attachments(self,page=1,per_page=20):
        return ZCRMModuleRelation.get_instance(self,"Attachments").get_attachments(page,per_page)

    def upload_attachment(self,file_path):
        if file_path is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Upload_Attachment',"file path must be given",'FILEPATH NOT PROVIDED',"FILEPATH")
        return ZCRMModuleRelation.get_instance(self, "Attachments").upload_attachment(file_path)
    
    def upload_link_as_attachment(self,link_url):
        if link_url is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Upload_Link_As_Attachment',"link URL must be given",'URL_LINK NOT PROVIDED',"URL_LINK")
        return ZCRMModuleRelation.get_instance(self, "Attachments").upload_link_as_attachment(link_url)
    
    def download_attachment(self,attachment_id):
        if attachment_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Download_Attachment',"attachment id must be given",'ID DOES NOT PROVIDED',"ID")
        return ZCRMModuleRelation.get_instance(self, "Attachments").download_attachment(attachment_id)
    
    def delete_attachment(self,attachment_id):
        if attachment_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Delete_Attachment',"attachment id must be given",'ID DOES NOT PROVIDED',"ID")
        return ZCRMModuleRelation.get_instance(self, "Attachments").delete_attachment(attachment_id)
    
    def upload_photo(self,file_path):
        if file_path is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Upload_Photo',"file path must be given",'FILEPATH NOT PROVIDED',"FILEPATH")
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        return EntityAPIHandler.get_instance(self).upload_photo(file_path)
    
    def download_photo(self):
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        return EntityAPIHandler.get_instance(self).download_photo()
    
    def delete_photo(self):
        try:
            from .Handler import EntityAPIHandler
        except ImportError:
            from Handler import EntityAPIHandler
        return EntityAPIHandler.get_instance(self).delete_photo()
    
    def add_relation(self,junction_record):
        if junction_record is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('add_relation',"junction_record must be given",'JUNCTION RECORD NOT PROVIDED',"JUNCTION RECORD")
        return ZCRMModuleRelation.get_instance(self,junction_record).add_relation()
    
    def remove_relation(self,junction_record):
        if junction_record is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('remove_relation',"junction_record must be given",'JUNCTION RECORD NOT PROVIDED',"JUNCTION RECORD")
        return ZCRMModuleRelation.get_instance(self,junction_record).remove_relation()
    
    def add_note(self,note_ins):
        if note_ins is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('add_note',"note instance must be given",'NOTE INSTANCE NOT PROVIDED',"NOTE INSTANCE")
        return ZCRMModuleRelation.get_instance(self,"Notes").add_note(note_ins)
    
    def update_note(self,note_ins):
        if note_ins is None or note_ins.id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('update_note',"note instance and note id must be given",'NOTE INSTANCE OR NOTE ID NOT PROVIDED',"NOTE INSTANCE OR ID")
        return ZCRMModuleRelation.get_instance(self,"Notes").update_note(note_ins)
    
    def delete_note(self,note_ins):
        if note_ins is None or note_ins.id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('delete_note',"note instance and note id must be given",'NOTE INSTANCE OR NOTE ID NOT PROVIDED',"NOTE INSTANCE OR ID")
        return ZCRMModuleRelation.get_instance(self,"Notes").delete_note(note_ins)
    
    def get_notes(self,sort_by=None,sort_order=None,page=1,per_page=20):
        return ZCRMModuleRelation.get_instance(self,"Notes").get_notes(sort_by,sort_order,page,per_page)

    def get_relatedlist_records(self,relatedlist_api_name,sort_by=None,sort_order=None,page=1,per_page=20):
        return ZCRMModuleRelation.get_instance(self,relatedlist_api_name).get_records(sort_by,sort_order,page,per_page)

    def add_tags(self, tagnames):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance().add_tags(self, tagnames)

    def get_blueprint(self):
        if self.entity_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('GET_BLUEPRINT', "id should be set for the record", 'ID IS NOT PROVIDED', "ID")
        return ZCRMBlueprint.get_instance(self,self.module_api_name,self.entity_id).get_blueprint()

    def remove_tags(self,tagnames):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance().remove_tags(self, tagnames)

    def update_blueprint(self):
        if self.entity_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('UPDATE_BLUEPRINT', "id should be set for the record", 'ID IS NOT PROVIDED', "ID")
        if self.blueprint_transition_id is None:
            raise Exception("Transition ID is mandatory")
        return ZCRMBlueprint.get_instance(self, self.module_api_name, self.entity_id).update_blueprint()

class ZCRMInventoryLineItem(object):
    '''
    This class is to deal with Zoho CRM Inventory line items
    '''
    def __init__(self,param):
        if(isinstance(param, ZCRMRecord)):
            self.product=param
            self.id=None
        else:
            self.id=param
            self.product=None
        
        self.list_price=None
        self.quantity=None
        self.description=None
        self.total=None
        self.discount=None
        self.discount_percentage=None
        self.total_after_discount=None
        self.tax_amount=None
        self.net_total=None
        self.delete_flag = False
        self.line_tax = list()
        
    @staticmethod
    def get_instance(param):
        return ZCRMInventoryLineItem(param)

class ZCRMTag(object):
    def __init__(self, tag_id, name):
        self.id = tag_id
        self.name = name
        self.module_apiname = None
        self.created_by = None
        self.modified_by = None
        self.created_time = None
        self.modified_time = None
        self.count = None

    @staticmethod
    def get_instance(tag_id=None, name=None):
        return ZCRMTag(tag_id, name)

    def delete(self):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance().delete(self.id)

    def merge(self, merge_tag):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance().merge(self.id, merge_tag.id)

    def update(self):
        try:
            from .Handler import TagAPIHandler
        except ImportError:
            from Handler import TagAPIHandler
        return TagAPIHandler.get_instance().update(self)

class ZCRMTax(object):
    
    def __init__(self,name):
        self.name=name
        self.percentage=None
        self.value=None
        
    @staticmethod
    def get_instance(name):
        return ZCRMTax(name)


class ZCRMOrgTax(object):

    def __init__(self, org_tax_id, name):
        self.id = org_tax_id
        self.name = name
        self.display_label = None
        self.value = None
        self.sequence_number = None

    @staticmethod
    def get_instance(org_tax_id=None, name=None):
        return ZCRMOrgTax(org_tax_id, name)

class ZCRMEventParticipant(object):
    def __init__(self,participant_type,participant_id):
        self.id = participant_id
        self.type = participant_type
        self.email=None
        self.name=None
        self.is_invited=None
        self.status=None 
    @staticmethod
    def get_instance(participant_type,participant_id):
        return ZCRMEventParticipant(participant_type,participant_id)
    
class ZCRMPriceBookPricing(object):
    def __init__(self,price_book_id):
        self.id=price_book_id
        self.to_range=None
        self.from_range=None
        self.discount=None
    @staticmethod
    def get_instance(price_book_id):
        return ZCRMPriceBookPricing(price_book_id)

class ZCRMUser(object):
    def __init__(self,user_id,name):
        self.id=user_id
        self.name=name
        self.signature=None
        self.country=None
        self.role=None
        self.customize_info=None
        self.city=None
        self.name_format=None
        self.language=None
        self.locale=None
        self.is_personal_account=None
        self.default_tab_group=None
        self.street=None
        self.alias=None
        self.theme=None
        self.state=None
        self.country_locale=None
        self.fax=None
        self.first_name=None
        self.email=None
        self.zip=None
        self.decimal_separator=None
        self.website=None
        self.time_format=None
        self.profile=None
        self.mobile=None
        self.last_name=None
        self.time_zone=None
        self.zuid=None
        self.is_confirm=None
        self.full_name=None
        self.phone=None
        self.dob=None
        self.date_format=None
        self.status=None
        self.created_by=None;
        self.modified_by=None;
        self.territories=None;
        self.reporting_to=None;
        self.is_online=None;
        self.currency=None;
        self.created_time=None;
        self.modified_time=None;
        self.field_apiname_vs_value=dict();
    @staticmethod
    def get_instance(user_id=None,name=None):
        return ZCRMUser(user_id,name)
    defaultKeys=["Currency","Modified_Time","created_time","territories","reporting_to","Isonline","created_by","Modified_By","country","id","name","role","customize_info","city","signature","name_format","language","locale","personal_account","default_tab_group","alias","street","theme","state","country_locale","fax","first_name","email","zip","decimal_separator","website","time_format","profile","mobile","last_name","time_zone","zuid","confirm","full_name","phone","dob","date_format","status"]
    
class ZCRMUserCustomizeInfo(object):
    def __init__(self):
        self.notes_desc=None
        self.is_to_show_right_panel=None
        self.is_bc_view=None
        self.is_to_show_home=None
        self.is_to_show_detail_view=None
        self.unpin_recent_item=None
    @staticmethod
    def get_instance():
        return ZCRMUserCustomizeInfo()
    
class ZCRMUserTheme(object):
    
    def __init__(self):
        self.normal_tab_font_color=None
        self.normal_tab_background=None
        self.selected_tab_font_color=None
        self.selected_tab_background=None
        
    @staticmethod
    def get_instance():
        return ZCRMUserTheme()
    
class ZCRMRole(object):
    def __init__(self,role_id,role_name):
        self.name=role_name
        self.id=role_id
        self.reporting_to=None
        self.label=None
        self.is_admin=None
        self.forecast_manager = None
        self.share_with_peers = None
    
    @staticmethod
    def get_instance(role_id,role_name=None):
        return ZCRMRole(role_id,role_name)
    
        
class ZCRMLayout(object):
    
    def __init__(self,layout_id):
        self.id=layout_id
        self.name=None
        self.created_time=None
        self.modified_time=None
        self.is_visible=None
        self.modified_by=None
        self.accessible_profiles=None
        self.created_by=None
        self.sections=None
        self.status=None
        self.convert_mapping=dict()
        
    @staticmethod
    def get_instance(layout_id):
        return ZCRMLayout(layout_id)
    
class ZCRMAttachment(object):
    def __init__(self,parent_record,attachment_id):
        self.id=attachment_id
        self.parent_record=parent_record
        self.file_name=None
        self.file_type=None
        self.size=None
        self.owner=None
        self.created_by=None
        self.created_time=None
        self.modified_by=None
        self.modified_time=None
        self.parent_module=None
        self.attachment_type=None
        self.parent_name=None
        self.parent_id=None
    @staticmethod
    def get_instance(parent_record,attachment_id=None):
        return ZCRMAttachment(parent_record,attachment_id)

class ZCRMCustomView(object):
    
    def __init__(self,custom_view_id,module_apiname):
        self.id=custom_view_id
        self.module_api_name=module_apiname
        self.display_value=None
        self.is_default=None
        self.name=None
        self.system_name=None
        self.sort_by=None
        self.category=None
        self.fields=list()
        self.favorite=None
        self.sort_order=None
        self.criteria_pattern=None
        self.criteria=None
        self.categories=list()
        self.is_off_line=None
        
    @staticmethod
    def get_instance(custom_view_id,module_apiname=None):
        return ZCRMCustomView(custom_view_id,module_apiname)
    
class ZCRMCustomViewCategory(object):
    
    def __init__(self):
        self.display_value=None
        self.actual_value=None
    @staticmethod
    def get_instance():
        return ZCRMCustomViewCategory()
    
class ZCRMCustomViewCriteria(object):
    
    def __init__(self):
        self.comparator=None
        self.field=None
        self.value=None
        
    @staticmethod
    def get_instance():
        return ZCRMCustomViewCriteria()
    
class ZCRMField(object):
    
    def __init__(self,api_name):
        self.api_name=api_name
        self.is_custom_field=None
        self.lookup_field=None
        self.convert_mapping=None
        self.is_visible=None
        self.field_label=None
        self.length=None
        self.created_source=None
        self.default_value=None
        self.is_mandatory=None
        self.sequence_number=None
        self.is_read_only=None
        self.is_unique_field=None
        self.is_case_sensitive=None
        self.data_type=None
        self.is_formula_field=None
        self.is_currency_field=None
        self.id=None
        self.picklist_values=list()
        self.is_auto_number=None
        self.is_business_card_supported=None
        self.field_layout_permissions=None
        self.decimal_place=None
        self.precision=None
        self.rounding_option=None
        self.formula_return_type=None
        self.formula_expression=None
        self.prefix=None
        self.suffix=None
        self.start_number=None
        self.json_type=None
        self.is_webhook = None
        self.crypt = None
        self.tooltip = None
        self.created_source = None
        self.display_label = None
        self.association_details = None
        self.column_name = None
        self.type = None
        self.is_history_tracking = None
        self.transition_sequence = None
        self.is_system_mandatory = None
        self.related_details = None
        self.personality_name = None
        self.layouts = None
        self.criteria = None

    @staticmethod
    def get_instance(api_name):
        return ZCRMField(api_name)
    
class ZCRMJunctionRecord(object):
    
    def __init__(self,api_name,record_id):
        self.id=record_id
        self.api_name=api_name
        self.related_data=dict()
        
    @staticmethod
    def get_instance(api_name,record_id):
        return ZCRMJunctionRecord(api_name,record_id)
    
    def set_related_data(self,key,value):
        self.related_data[key]=value
        
    def get_related_data(self):
        return self.related_data
        
class ZCRMLeadConvertMapping(object):
    
    def __init__(self,name,converted_id):
        self.id=converted_id
        self.name=name
        self.fields=list()
    @staticmethod
    def get_instance(name,converted_id):
        return ZCRMLeadConvertMapping(name,converted_id)
    
class ZCRMLeadConvertMappingField(object):
    
    def __init__(self,api_name,field_id):
        self.id=field_id
        self.api_name=api_name
        self.field_label=None
        self.is_required=None
        
    @staticmethod
    def get_instance(api_name,field_id):
        return ZCRMLeadConvertMappingField(api_name,field_id)
    
class ZCRMLookupField(object):
    
    def __init__(self,apiName):
        self.api_name=apiName
        self.display_label=None
        self.module=None
        self.id=None
        self.field_label = None
        self.type = None

    @staticmethod
    def get_instance(api_name):
        return ZCRMLookupField(api_name)

class ZCRMBlueprintRelatedModule(object):

    def __init__(self,id):
        self.id = id
        self.display_label = None
        self.module_name = None

    @staticmethod
    def get_instance(id):
        return ZCRMBlueprintRelatedModule(id)

class ZCRMModuleRelatedList(object):
     
    def __init__(self,api_name):
        self.api_name=api_name
        self.module=None
        self.display_label=None
        self.is_visible=None
        self.name=None
        self.id=None
        self.href=None
        self.type=None 
    
    @staticmethod
    def get_instance(api_name):
        return ZCRMModuleRelatedList(api_name)
    
    def set_relatedlist_properties(self,relatedlist_prop):
        self.id=relatedlist_prop['id']
        self.module=relatedlist_prop['module']
        self.display_label=relatedlist_prop['display_label']
        self.name=relatedlist_prop['name']
        self.type=relatedlist_prop['type']
        self.href=relatedlist_prop['href']
        self.is_visible=bool(relatedlist_prop['visible']) if 'visible' in relatedlist_prop else None
        return self
    
class ZCRMModuleRelation(object):
    
    def __init__(self,parentmodule_apiname_or_parentrecord,related_list_apiname_or_junction_record):
        if isinstance(parentmodule_apiname_or_parentrecord, ZCRMRecord):
            self.parent_record=parentmodule_apiname_or_parentrecord
            self.parent_module_api_name=None
        else:
            self.parent_module_api_name=parentmodule_apiname_or_parentrecord
            self.parent_record=None
        
        if isinstance(related_list_apiname_or_junction_record, ZCRMJunctionRecord):
            self.junction_record=related_list_apiname_or_junction_record
            self.api_name=None
        else:
            self.api_name=related_list_apiname_or_junction_record
            self.junction_record=None
            
        self.label=None
        self.id=None
        self.is_visible=None
        
    @staticmethod
    def get_instance(parentmodule_apiname_or_parentrecord,related_list_apiname_or_junction_record):
        return ZCRMModuleRelation(parentmodule_apiname_or_parentrecord,related_list_apiname_or_junction_record)
    
    def get_records(self,sort_by_field=None,sort_order=None,page=1,per_page=20):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).get_records(sort_by_field,sort_order,page,per_page)
    def upload_attachment(self,file_path):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).upload_attachment(file_path)
    def upload_link_as_attachment(self,link_url):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).upload_link_as_attachment(link_url)
    def download_attachment(self,attachment_id):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).download_attachment(attachment_id)
    def delete_attachment(self,attachment_id):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).delete_attachment(attachment_id)
    def add_relation(self):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self.junction_record).add_relation()
    def remove_relation(self):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self.junction_record).remove_relation()
    def add_note(self,zcrm_note_ins):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).add_note(zcrm_note_ins)
    def update_note(self,zcrm_note_ins):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).update_note(zcrm_note_ins)
    def delete_note(self,zcrm_note_ins):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).delete_note(zcrm_note_ins)
    def get_notes(self,sort_by,sort_order,page,per_page):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).get_notes(sort_by,sort_order,page,per_page)
    def get_attachments(self,page,per_page):
        try:
            from .Handler import RelatedListAPIHandler
        except ImportError:
            from Handler import RelatedListAPIHandler
        return RelatedListAPIHandler.get_instance(self.parent_record,self).get_attachments(page,per_page)
    
class ZCRMNote(object):
    
    def __init__(self,parent_record,note_id):
        self.id=note_id
        self.parent_record=parent_record
        self.title=None
        self.content=None
        self.owner=None
        self.created_by=None
        self.created_time=None
        self.modified_by=None
        self.modified_time=None
        self.attachments=None
        self.size=None
        self.is_voice_note=None
        self.parent_module=None
        self.parent_name=None
        self.parent_id=None
        
    @staticmethod
    def get_instance(parent_record,note_id=None):
        return ZCRMNote(parent_record,note_id)

    def get_attachments(self, page=1, per_page=20):
        return ZCRMModuleRelation.get_instance(ZCRMRecord.get_instance('Notes', self.id), "Attachments").get_attachments(page, per_page)

    def upload_attachment(self, file_path):
        if file_path is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Upload_Attachment', "file path must be given", 'FILEPATH NOT PROVIDED',
                                       "FILEPATH")
        return ZCRMModuleRelation.get_instance(ZCRMRecord.get_instance('Notes', self.id), "Attachments").upload_attachment(file_path)

    def download_attachment(self, attachment_id):
        if attachment_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Download_Attachment', "attachment id must be given", 'ID DOES NOT PROVIDED',
                                       "ID")
        return ZCRMModuleRelation.get_instance(ZCRMRecord.get_instance('Notes', self.id), "Attachments").download_attachment(attachment_id)

    def delete_attachment(self, attachment_id):
        if attachment_id is None:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            CommonUtil.raise_exception('Delete_Attachment', "attachment id must be given", 'ID DOES NOT PROVIDED', "ID")
        return ZCRMModuleRelation.get_instance(ZCRMRecord.get_instance('Notes', self.id), "Attachments").delete_attachment(attachment_id)
    
class ZCRMPermission(object):
    
    def __init__(self,permission_name,permission_id):
        self.id=permission_id;
        self.name=permission_name;
        self.display_label=None;
        self.module=None;
        self.is_enabled=None;
        
    @staticmethod
    def get_instance(permission_name,permission_id):
        return ZCRMPermission(permission_name,permission_id)
    
class ZCRMPickListValue(object):
    
    def __init__(self):
        self.display_value=None
        self.sequence_number=None
        self.actual_value=None
        self.maps=None
    @staticmethod
    def get_instance():
        return ZCRMPickListValue()
    
class ZCRMProfile(object):
    
    def __init__(self,profile_id,profile_name):
        self.name=profile_name
        self.id=profile_id
        self.is_default=None
        self.created_time=None
        self.modified_time=None
        self.modified_by=None
        self.description=None
        self.created_by=None
        self.category=None
        self.permissions=list()
        self.sections=list()
    
    @staticmethod
    def get_instance(profile_id,profile_name=None):
        return ZCRMProfile(profile_id,profile_name)
    
class ZCRMProfileCategory(object):
    
    def __init__(self,profile_category_name):
        self.name=profile_category_name
        self.module=None
        self.display_label=None
        self.permission_ids=list()
        
    @staticmethod
    def get_instance(profile_category_name):
        return ZCRMProfileCategory(profile_category_name)
    
class ZCRMProfileSection(object):
    
    def __init__(self,section_name):
        self.name=section_name
        self.categories=list()
        
    @staticmethod
    def get_instance(section_name):
        return ZCRMProfileSection(section_name)
    
class ZCRMRelatedListProperties(object):
    
    def __init__(self):
        self.sort_by=None
        self.sort_order=None
        self.fields=None
        
    @staticmethod
    def get_instance():
        return ZCRMRelatedListProperties()
    
class ZCRMSection(object):
    
    def __init__(self,section_name):
        self.name=section_name
        self.display_name=None
        self.column_count=None
        self.sequence_number=None
        self.fields=None
        
    @staticmethod
    def get_instance(section_name):
        return ZCRMSection(section_name)
    

class ZCRMTrashRecord(object):
    
    def __init__(self,entity_type,entity_id):
        self.id = entity_id
        self.type=entity_type
        self.display_name=None
        self.deleted_time=None
        self.created_by=None
        self.deleted_by=None
        
    @staticmethod
    def get_instance(entity_type,entity_id=None):
        return ZCRMTrashRecord(entity_type,entity_id)


try:
    from .Handler import VariableAPIHandler
except Exception:
    from Handler import VariableAPIHandler

class ZCRMVariable(object):
    def __init__(self):
        self.id = None
        self.name = None
        self.api_name = None
        self.type = None
        self.value = None
        self.variable_group = dict()
        self.description = None

    @staticmethod
    def get_instance():
        return ZCRMVariable()

    def get_variable(self, group):
        handler_ins = VariableAPIHandler.get_instance()
        handler_ins.variable = self
        return handler_ins.get_variable(group)

    def update_variable(self):
        handler_ins = VariableAPIHandler.get_instance()
        handler_ins.variable = self
        return handler_ins.update_variable()

    def delete_variable(self):
        handler_ins = VariableAPIHandler.get_instance()
        handler_ins.variable = self
        return handler_ins.delete_variable()


try:
    from .Handler import VariableGroupAPIHandler
except Exception:
    from Handler import VariableGroupAPIHandler


class ZCRMVariableGroup(object):
    def __init__(self):
        self.id = None
        self.name = None
        self.api_name = None
        self.display_label = None
        self.description = None

    @staticmethod
    def get_instance():
        return ZCRMVariableGroup()

    def get_variable_group(self):
        handler_ins = VariableGroupAPIHandler.get_instance()
        handler_ins.variable_group = self
        return handler_ins.get_variable_group()

class ZCRMBlueprint(object):
    def __init__(self,zcrmrecord,entity_type,entity_id):
        self.zcrmrecord = zcrmrecord
        self.entity_type = entity_type
        self.entity_id = entity_id
        self.transitions = list()
        self.processinfodata = dict()

    @staticmethod
    def get_instance(zcmrecord,entity_type,entity_id):
        return ZCRMBlueprint(zcrmrecord=zcmrecord,entity_type=entity_type,entity_id=entity_id)

    def get_blueprint(self):
        try:
            from .Handler import BlueprintAPIHandler
        except Exception:
            from Handler import BlueprintAPIHandler
        try:
            apiresponse = BlueprintAPIHandler.get_instance(self.zcrmrecord,self.entity_type,self.entity_id).get()
            blueprint_json = apiresponse.response_json["blueprint"]
            self.set_processinfo_values(blueprint_json)
            transitionlist = blueprint_json["transitions"]
            for transitiondict in transitionlist:
                self.transitions.append(ZCRMBlueprintTransition.get_instance(self.entity_type).get_transition(transitiondict))
            apiresponse.data = self
            return apiresponse
        except Exception as e:
            raise e

    def update_blueprint(self):
        try:
            from .Handler import BlueprintAPIHandler
        except Exception:
            from Handler import BlueprintAPIHandler
        try:
            apiresponse = BlueprintAPIHandler.get_instance(self.zcrmrecord,self.entity_type,self.entity_id).update()
            return apiresponse
        except Exception as e:
            raise e

    def set_processinfo_values(self,responsedict):
        if "process_info" in responsedict:
            processinfodict = responsedict["process_info"]
            for key in processinfodict:
                value = processinfodict[key]
                self.processinfodata[key] = value;
        return self.processinfodata

    def get_processinfo_value(self,key):
        return self.processinfodata[key] if key in self.processinfodata else None

    def get_processinfo_values(self):
        return self.processinfodata

class ZCRMBlueprintTransition(object):
    def __init__(self,entity_name):
        self.entity_name = entity_name
        self.next_transitions = []
        self.transition_data = None
        self.transitionproperties = dict()
        self.fields = []
        self.attachments = list()
        self.checklists = list()

    @staticmethod
    def get_instance(entity_type):
        return ZCRMBlueprintTransition(entity_type)

    def get_transition_value(self,key):
        return self.transitionproperties[key] if key in self.transitionproperties else None

    def get_transition_values(self):
        return self.transitionproperties

    def get_transition(self,transitiondict):
            for key in transitiondict:
                value = transitiondict[key]
                if key == 'next_transitions' and 'next_transitions' in transitiondict:
                    transitiondetails = transitiondict['next_transitions']
                    for eachtransition in transitiondetails:
                        self.next_transitions.append(ZCRMTransition.get_instance(eachtransition['id'], eachtransition['name']))
                elif key == 'fields' and 'fields' in transitiondict:
                    fieldslist = transitiondict['fields']
                    self.get_fields(fieldslist)
                elif key =='data' and 'data' in transitiondict:
                    relatedlist_instance = ZCRMTransitionRelatedData.get_instance()
                    datadict = transitiondict['data']
                    for data_key in datadict:
                        if data_key == 'CheckLists' and 'CheckLists' in datadict:
                            self.checklists.append(ZCRMTransitionCheckList.get_instance(datadict['CheckLists']['title'] if 'title' in datadict['CheckLists'] else None).get_checklist(datadict['CheckLists']))
                        elif data_key == 'Attachments' and 'Attachments' in datadict:
                            for attachment in datadict['Attachments']:
                                self.attachments.append(ZCRMTransitionAttachment.get_instance(attachment['name'] if 'name' in attachment else None).set_attachment_values(attachment))
                        else:
                            relatedlist_instance.populate_data_values(data_key, datadict[data_key])
                else:
                    self.transitionproperties[key] = value
            self.transition_data = (relatedlist_instance)
            return self

    def get_fields(self, fieldslist):
        try:
            from .Operations import ZCRMField,ZCRMModule
            from .Handler import ModuleAPIHandler
        except ImportError:
            from Operations import ZCRMField,ZCRMModule
            from Handler import ModuleAPIHandler
        module_instance = ZCRMModule.get_instance(self.entity_name)
        handler_instance = ModuleAPIHandler.get_instance(module_instance)
        for field in fieldslist:
            if 'api_name' not in field:
                field['api_name'] = None
            self.fields.append(handler_instance.get_zcrmfield(field))

class ZCRMTransition(object):
    def __init__(self,id,name):
        self.name = name
        self.id = id

    @staticmethod
    def get_instance(id, name):
        return ZCRMTransition(id,name)

class ZCRMTransitionCheckList(object):
    def __init__(self,title):
        self.title = title
        self.items = list()

    @staticmethod
    def get_instance(title):
        return ZCRMTransitionCheckList(title)

    def get_checklist(self, checklistdict):
        if 'items' in checklistdict:
            itemslist = checklistdict['items']
            for item in itemslist:
                new_item = {}
                for keys in item:
                    new_item[keys] = item[keys]
                self.items.append(new_item)
        return self

class ZCRMTransitionRelatedData(object):
    def __init__(self):
        self.values = dict()

    @staticmethod
    def get_instance():
        return ZCRMTransitionRelatedData()

    def get_value(self, related_list, key):
        related_list_dict = self.values[related_list]
        return related_list_dict[key] if key in related_list_dict else None

    def populate_data_values(self, name, relatedlist_dict):
        if isinstance(relatedlist_dict, dict):
            current_relatedlist_dict = dict()
            for keys in relatedlist_dict:
                current_relatedlist_dict[str(keys)] = relatedlist_dict[str(keys)]
            self.values[name] = current_relatedlist_dict
        else:
            self.values[name] = relatedlist_dict

    def get_data_value(self, key):
        return self.values[key]

    def get_data_values(self):
        return self.values

    def get_data_names(self):
        return self.values.keys()

    def get_data_keys(self,key):
        return (self.values[key]).keys() if isinstance(self.values[key], dict) else self.values[key]


class ZCRMTransitionAttachment(object):
    def __init__(self,name):
        self.filesize = None
        self.name = name
        self.file_id = None
        self.link_url = None

    @staticmethod
    def get_instance(name):
        return ZCRMTransitionAttachment(name)

    def set_attachment_values(self,attachmentdict):
        self.file_id = attachmentdict['$file_id'] if '$file_id' in attachmentdict else None
        self.link_url = attachmentdict['$link_url'] if '$link_url' in attachmentdict else None
        self.filesize = attachmentdict['fileSize'] if 'fileSize' in attachmentdict else None
        return self

class ZCRMTransitionCrypt(object):
    def __init__(self):
        self.mode = None
        self.column = None
        self.table = None
        self.status = None

    @staticmethod
    def get_instance():
        return ZCRMTransitionCrypt()


'''
Created on Jul 28, 2017

@author: sumanth-3058
'''
try:
    from .Utility import ZCRMConfigUtil, APIConstants,HTTPConnector
    from .CLException import ZCRMException
    from .Response import APIResponse, BulkAPIResponse,FileAPIResponse
except ImportError:
    from Utility import ZCRMConfigUtil, APIConstants,HTTPConnector
    from CLException import ZCRMException
    from Response import APIResponse, BulkAPIResponse,FileAPIResponse
class APIRequest(object):
    '''
    This class is to wrap the API request related stuff like request params,headers,body,..etc
    '''
    def __init__(self, api_handler_ins):
        '''
        Constructor
        '''
        self.construct_api_url()
        self.url+=api_handler_ins.request_url_path
        if(not self.url.startswith("http")):
            self.url="https://"+self.url
        self.request_body=api_handler_ins.request_body
        self.request_headers=api_handler_ins.request_headers
        self.request_params=api_handler_ins.request_params
        self.request_method=api_handler_ins.request_method
        self.request_api_key=api_handler_ins.request_api_key
    
    def construct_api_url(self):
        hit_sand_box=ZCRMConfigUtil.config_prop_dict['sandbox']
        base_url=ZCRMConfigUtil.get_api_base_url().replace('www','sandbox') if hit_sand_box.lower()=='true' else ZCRMConfigUtil.get_api_base_url()
        self.url=base_url+"/crm/"+ZCRMConfigUtil.get_api_version()+"/"
    
    def authenticate_request(self):
        accessToken=ZCRMConfigUtil.get_instance().get_access_token()
        if(self.request_headers==None):
            self.request_headers={APIConstants.AUTHORIZATION:APIConstants.OAUTH_HEADER_PREFIX+accessToken}
        else:
            self.request_headers[APIConstants.AUTHORIZATION]=APIConstants.OAUTH_HEADER_PREFIX+accessToken
        self.request_headers['User-Agent']='ZohoCRM Python SDK'
    def get_api_response(self):
        try:
            self.authenticate_request()
            connector=HTTPConnector.get_instance(self.url, self.request_params, self.request_headers, self.request_body, self.request_method, self.request_api_key, False)
            response=connector.trigger_request()
            return APIResponse(response,response.status_code,self.url,self.request_api_key)
        except ZCRMException as ex:
            raise ex
        except Exception as ex:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            import traceback
            CommonUtil.raise_exception(self.url,ex.__str__(),traceback.format_stack())
        
    def get_bulk_api_response(self):
        try:
            self.authenticate_request()
            connector=HTTPConnector.get_instance(self.url, self.request_params, self.request_headers, self.request_body, self.request_method, self.request_api_key, True)
            response=connector.trigger_request()
            return BulkAPIResponse(response,response.status_code,self.url,self.request_api_key,self.request_method)
        except ZCRMException as ex:
            raise ex
        except Exception as ex:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            import traceback
            CommonUtil.raise_exception(self.url,ex.__str__(),traceback.format_stack())
        
    def upload_attachment(self,file_path,upload_file=True):
        try:
            self.authenticate_request()
            if upload_file:
                files = {'file': open(file_path,'rb')}
            else:
                files=self.request_body
            connector=HTTPConnector.get_instance(self.url, self.request_params, self.request_headers, self.request_body, self.request_method, self.request_api_key, True)
            connector.set_file(files)
            response=connector.trigger_request()
            return APIResponse(response,response.status_code,self.url,self.request_api_key)
        except ZCRMException as ex:
            raise ex
        except Exception as ex:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            import traceback
            CommonUtil.raise_exception(self.url,ex.__str__(),traceback.format_stack())
    def upload_link_as_attachment(self):
        try:
            self.authenticate_request()
            files = {'file': ''}
            connector=HTTPConnector.get_instance(self.url, self.request_params, self.request_headers, self.request_body, self.request_method, self.request_api_key, True)
            connector.set_file(files)
            response=connector.trigger_request()
            return APIResponse(response,response.status_code,self.url,self.request_api_key)
        except ZCRMException as ex:
            raise ex
        except Exception as ex:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            import traceback
            CommonUtil.raise_exception(self.url,ex.__str__(),traceback.format_stack())
      
    def download_attachment(self):
        try:
            self.authenticate_request()
            connector=HTTPConnector.get_instance(self.url, self.request_params, self.request_headers, self.request_body, self.request_method, self.request_api_key, True)
            response=connector.trigger_request()
            if response.status_code==APIConstants.RESPONSECODE_OK:
                file_res= FileAPIResponse(response,response.status_code,self.url)
                file_res.status=APIConstants.STATUS_SUCCESS
                content_disp=response.headers['Content-Disposition']
                start_index=content_disp.rindex("'")
                file_res.file_name=content_disp[start_index+1:]
                file_res.response_headers=response.headers
                return file_res
            elif(response.status_code==APIConstants.RESPONSECODE_NO_CONTENT):
                errorMsg=APIConstants.INVALID_DATA+"-"+APIConstants.INVALID_ID_MSG
                exception=ZCRMException(self.url,response.status_code,errorMsg,APIConstants.NO_CONTENT,None,errorMsg)
                exception.message = exception.__str__()
                raise exception
            else:
                responseJSON=response.json()
                exception=ZCRMException(self.url,response.status_code,responseJSON[APIConstants.MESSAGE],responseJSON[APIConstants.CODE],responseJSON[APIConstants.DETAILS],responseJSON[APIConstants.MESSAGE])
                exception.message = exception.__str__()
                raise exception
        except ZCRMException as ex:
            raise ex
        except Exception as ex:
            try:
                from .Utility import CommonUtil
            except ImportError:
                from Utility import CommonUtil
            import traceback
            CommonUtil.raise_exception(self.url,ex.__str__(),traceback.format_stack())


'''
Created on Aug 16, 2017

@author: sumanth-3058
'''
try:
    from .OAuthUtility import OAuthLogger,ZohoOAuthConstants,ZohoOAuthException,ZohoOAuthHTTPConnector,ZohoOAuthParams
    from .Persistence import ZohoOAuthPersistenceHandler,ZohoOAuthPersistenceFileHandler
except ImportError:
    from OAuthUtility import OAuthLogger,ZohoOAuthConstants,ZohoOAuthException,ZohoOAuthHTTPConnector,ZohoOAuthParams
    from Persistence import ZohoOAuthPersistenceHandler,ZohoOAuthPersistenceFileHandler
import logging
class ZohoOAuth(object):
    '''
    This class is to load oauth configurations and provide OAuth request URIs
    '''
    configProperties={}
    #iamURL='https://accounts.zoho.com'
    
    def __init__(self):
        '''
        Constructor
        '''
    @staticmethod
    def initialize(config_dict = None):
        try:
            ZohoOAuth.set_config_values(config_dict)
            if((ZohoOAuthConstants.TOKEN_PERSISTENCE_PATH not in ZohoOAuth.configProperties or ZohoOAuth.configProperties[ZohoOAuthConstants.TOKEN_PERSISTENCE_PATH] =="") and \
                    (ZohoOAuthConstants.CUSTOM_PERSISTENCE_HANDLER_PATH not in ZohoOAuth.configProperties or ZohoOAuth.configProperties[ZohoOAuthConstants.CUSTOM_PERSISTENCE_HANDLER_PATH] == "")):
                if(ZohoOAuthConstants.DATABASE_PORT not in ZohoOAuth.configProperties or ZohoOAuth.configProperties[ZohoOAuthConstants.DATABASE_PORT]==""):
                    ZohoOAuth.configProperties[ZohoOAuthConstants.DATABASE_PORT]="3306"
                if(ZohoOAuthConstants.DATABASE_USERNAME not in ZohoOAuth.configProperties or ZohoOAuth.configProperties[ZohoOAuthConstants.DATABASE_USERNAME]==""):
                    ZohoOAuth.configProperties[ZohoOAuthConstants.DATABASE_USERNAME]="root"
                if(ZohoOAuthConstants.DATABASE_PASSWORD not in ZohoOAuth.configProperties or ZohoOAuth.configProperties[ZohoOAuthConstants.DATABASE_PASSWORD]==""):
                    ZohoOAuth.configProperties[ZohoOAuthConstants.DATABASE_PASSWORD]=""
            oAuthParams=ZohoOAuthParams.get_instance(ZohoOAuth.configProperties[ZohoOAuthConstants.CLIENT_ID], ZohoOAuth.configProperties[ZohoOAuthConstants.CLIENT_SECRET], ZohoOAuth.configProperties[ZohoOAuthConstants.REDIRECT_URL])
            ZohoOAuthClient.get_instance(oAuthParams)
        except Exception as ex:
            OAuthLogger.add_log('Exception occured while reading oauth configurations',logging.ERROR,ex)
            raise ex
    
    @staticmethod
    def set_config_values(config_dict):
        config_keys = [ZohoOAuthConstants.CLIENT_ID, ZohoOAuthConstants.CLIENT_SECRET, ZohoOAuthConstants.REDIRECT_URL, ZohoOAuthConstants.ACCESS_TYPE
			, ZohoOAuthConstants.IAM_URL, ZohoOAuthConstants.TOKEN_PERSISTENCE_PATH, ZohoOAuthConstants.DATABASE_PORT
			, ZohoOAuthConstants.DATABASE_PASSWORD, ZohoOAuthConstants.DATABASE_USERNAME, ZohoOAuthConstants.CUSTOM_PERSISTENCE_HANDLER_PATH, ZohoOAuthConstants.CUSTOM_PERSISTENCE_HANDLER_CLASS]
        if(ZohoOAuthConstants.ACCESS_TYPE not in config_dict or config_dict[ZohoOAuthConstants.ACCESS_TYPE] is None):
            ZohoOAuth.configProperties[ZohoOAuthConstants.ACCESS_TYPE] = "offline"
        if(ZohoOAuthConstants.IAM_URL not in config_dict or config_dict[ZohoOAuthConstants.IAM_URL] == ""):
            ZohoOAuth.configProperties[ZohoOAuthConstants.IAM_URL] = "https://accounts.zoho.com"
        for key in config_keys:
            if(key in config_dict and config_dict[key] !=""):
                ZohoOAuth.configProperties[key] = config_dict[key].strip()
    @staticmethod
    def get_file_content_as_dictionary(filePointer) :
        dictionary={}
        for line in filePointer:
            line=line.rstrip()
            keyValue=line.split("=")
            if(not keyValue[0].startswith('#')):
                dictionary[keyValue[0].strip()]=keyValue[1].strip()
        filePointer.close()
        return dictionary
    @staticmethod
    def get_grant_url():
        return (ZohoOAuth.configProperties[ZohoOAuthConstants.IAM_URL]+"/oauth/v2/auth")
    @staticmethod
    def get_token_url():
        return (ZohoOAuth.configProperties[ZohoOAuthConstants.IAM_URL]+"/oauth/v2/token")
    @staticmethod
    def get_refresh_token_url():
        return (ZohoOAuth.configProperties[ZohoOAuthConstants.IAM_URL]+"/oauth/v2/token")
    @staticmethod
    def get_revoke_token_url():
        return (ZohoOAuth.configProperties[ZohoOAuthConstants.IAM_URL]+"/oauth/v2/token/revoke")
    @staticmethod
    def get_user_info_url():
        return (ZohoOAuth.configProperties[ZohoOAuthConstants.IAM_URL]+"/oauth/user/info")
    @staticmethod
    def get_client_instance():
        oauth_client_ins=ZohoOAuthClient.get_instance()
        if(oauth_client_ins is None):
            raise ZohoOAuthException('ZohoOAuth.initialize() must be called before this')
        return oauth_client_ins
    @staticmethod
    def get_persistence_instance():
        if((ZohoOAuthConstants.TOKEN_PERSISTENCE_PATH not in ZohoOAuth.configProperties or ZohoOAuth.configProperties[ZohoOAuthConstants.TOKEN_PERSISTENCE_PATH]=="") and \
                (ZohoOAuthConstants.CUSTOM_PERSISTENCE_HANDLER_PATH not in ZohoOAuth.configProperties or ZohoOAuth.configProperties[ZohoOAuthConstants.CUSTOM_PERSISTENCE_HANDLER_PATH] == "")):
            return ZohoOAuthPersistenceHandler()
        elif((ZohoOAuthConstants.TOKEN_PERSISTENCE_PATH in ZohoOAuth.configProperties) and ZohoOAuth.configProperties[ZohoOAuthConstants.TOKEN_PERSISTENCE_PATH] != ""):
            return ZohoOAuthPersistenceFileHandler()
        else:
            try:
                from sys import path
                import importlib
                custompersistence_handler = ZohoOAuth.configProperties[ZohoOAuthConstants.CUSTOM_PERSISTENCE_HANDLER_PATH]
                custompersistence_classname = ZohoOAuth.configProperties[ZohoOAuthConstants.CUSTOM_PERSISTENCE_HANDLER_CLASS]
                if custompersistence_classname == "":
                    raise ZohoOAuthException("Token Persistence Class must be given.")
                splitter = "/" if "/" in custompersistence_handler else "\\"
                custompersistence_directory = custompersistence_handler.strip(custompersistence_handler.split(splitter)[-1]).rstrip(splitter)
                path.append(custompersistence_directory)
                custompersistence_modulename = custompersistence_handler.split(splitter)[-1].rstrip(".py")
                try:
                    custompersistence_module = importlib.import_module(custompersistence_modulename)
                    custompersistence_class = getattr(custompersistence_module, custompersistence_classname)
                    custompersistence_instance = custompersistence_class()
                    return custompersistence_instance
                except Exception as e:
                    raise e
            except Exception as ex:
                OAuthLogger.add_log("Exception occured while fetching instance for Custom DB Persistence", logging.ERROR, ex)
                raise ex

class ZohoOAuthClient(object):
    '''
    This class is to generate oauth related tokens
    '''
    oAuthParams=None
    oAuthClientIns=None

    def __init__(self, oauthParams):
        '''
        Constructor
        '''
        ZohoOAuthClient.oAuthParams=oauthParams
        
    @staticmethod
    def get_instance(param=None):
        if(param!=None and ZohoOAuthClient.oAuthClientIns==None):
            ZohoOAuthClient.oAuthClientIns=ZohoOAuthClient(param)
        return ZohoOAuthClient.oAuthClientIns
            
    def get_access_token(self,userEmail):
        try:
            handler=ZohoOAuth.get_persistence_instance()
            oAuthTokens=handler.get_oauthtokens(userEmail)
            try:
                return oAuthTokens.get_access_token()
            except ZohoOAuthException as ex:
                OAuthLogger.add_log("Access token expired hence refreshing", logging.INFO, ex)
                oAuthTokens = self.refresh_access_token(oAuthTokens.refreshToken, userEmail)
                return oAuthTokens.accessToken
            except Exception as e:
                raise e
        except Exception as ex:
            OAuthLogger.add_log("Exception occurred while fetching oauthtokens", logging.ERROR, ex)
            raise ex
    def generate_access_token_from_refresh_token(self,refreshToken,userEmail):
        self.refresh_access_token(refreshToken, userEmail)
    def refresh_access_token(self,refreshToken,userEmail):
        if(refreshToken==None):
            raise ZohoOAuthException("Refresh token not provided!")
        try:
            connector=self.get_connector(ZohoOAuth.get_refresh_token_url())
            connector.add_http_request_params(ZohoOAuthConstants.GRANT_TYPE,ZohoOAuthConstants.GRANT_TYPE_REFRESH)
            connector.add_http_request_params(ZohoOAuthConstants.REFRESH_TOKEN,refreshToken)
            connector.set_http_request_method(ZohoOAuthConstants.REQUEST_METHOD_POST)
            response=connector.trigger_request()
            responseJSON=response.json()
            if(ZohoOAuthConstants.ACCESS_TOKEN in responseJSON):
                oAuthTokens=self.get_tokens_from_json(responseJSON)
                oAuthTokens.set_user_email(userEmail)
                oAuthTokens.refreshToken=refreshToken
                ZohoOAuth.get_persistence_instance().save_oauthtokens(oAuthTokens)
                return oAuthTokens
            
        except ZohoOAuthException as ex:
            OAuthLogger.add_log("Exception occured while refreshing oauthtoken",logging.ERROR,ex)
            raise ex
            
    def generate_access_token(self,grantToken):
        if(grantToken==None):
            raise ZohoOAuthException("Grant token not provided!")
        try:
            connector=self.get_connector(ZohoOAuth.get_token_url())
            connector.add_http_request_params(ZohoOAuthConstants.GRANT_TYPE,ZohoOAuthConstants.GRANT_TYPE_AUTH_CODE)
            connector.add_http_request_params(ZohoOAuthConstants.CODE,grantToken)
            connector.set_http_request_method(ZohoOAuthConstants.REQUEST_METHOD_POST)
            response=connector.trigger_request()
            responseJSON=response.json()
            if(ZohoOAuthConstants.ACCESS_TOKEN in responseJSON):
                oAuthTokens=self.get_tokens_from_json(responseJSON)
                oAuthTokens.set_user_email(self.get_user_email_from_iam(oAuthTokens.accessToken))
                ZohoOAuth.get_persistence_instance().save_oauthtokens(oAuthTokens)
                return oAuthTokens
            else:
                raise ZohoOAuthException("Exception occured while fetching accesstoken from Grant Token;Response is:"+str(responseJSON))
            
        except ZohoOAuthException as ex:
            OAuthLogger.add_log("Exception occured while generating access token",logging.ERROR,ex)
            raise ex
    
    def get_tokens_from_json(self,responseJson):
        expiresIn = responseJson[ZohoOAuthConstants.EXPIRES_IN] if 'expires_in_sec' in responseJson else (responseJson[ZohoOAuthConstants.EXPIRES_IN] * 1000)
        expiresIn=expiresIn+(ZohoOAuthTokens.get_current_time_in_millis())
        accessToken=responseJson[ZohoOAuthConstants.ACCESS_TOKEN]
        refreshToken=None
        if(ZohoOAuthConstants.REFRESH_TOKEN in responseJson):
            refreshToken=responseJson[ZohoOAuthConstants.REFRESH_TOKEN]
        oAuthTokens = ZohoOAuthTokens(refreshToken,accessToken,expiresIn)
        return oAuthTokens

    def get_connector(self,url):
        connector=ZohoOAuthHTTPConnector.get_instance(url,{})
        connector.add_http_request_params(ZohoOAuthConstants.CLIENT_ID, ZohoOAuthClient.oAuthParams.clientID)
        connector.add_http_request_params(ZohoOAuthConstants.CLIENT_SECRET, ZohoOAuthClient.oAuthParams.clientSecret)
        connector.add_http_request_params(ZohoOAuthConstants.REDIRECT_URL, ZohoOAuthClient.oAuthParams.redirectUri)
        return connector

    def get_user_email_from_iam(self,accessToken):
        header={ZohoOAuthConstants.AUTHORIZATION:(ZohoOAuthConstants.OAUTH_HEADER_PREFIX+accessToken)}
        connector=ZohoOAuthHTTPConnector.get_instance(ZohoOAuth.get_user_info_url(),None,header,None,ZohoOAuthConstants.REQUEST_METHOD_GET)
        response=connector.trigger_request()
        try:
            response_json = response.json()
            if 'Email' in response_json:
                return response.json()['Email']
        except ValueError as err:
            raise ZohoOAuthException('Exception while fetching User email from access token, Make sure AAAserver.profile.Read scope is included while generating the Grant token')

class ZohoOAuthTokens(object):
    '''
    This class is to encapsulate the OAuth tokens
    '''
    def __init__(self, refresh_token,access_token,expiry_time,user_email=None):
        '''
        Constructor
        '''
        self.refreshToken=refresh_token
        self.accessToken=access_token
        self.expiryTime=expiry_time
        self.userEmail=user_email
        
    def get_access_token(self):
        if(self.expiryTime-self.get_current_time_in_millis()) > 5000:
            return self.accessToken
        else:
            raise ZohoOAuthException("Access token got expired!")
    @staticmethod
    def get_current_time_in_millis():
        import time
        return int(round(time.time() * 1000))
    def set_user_email(self,userEmail):
        self.userEmail=userEmail

import sys
if sys.version > '3':
    from abc import ABC, abstractmethod
    class AbstractZohoOAuthPersistence(ABC):
        @abstractmethod
        def get_oauthtokens(self,user_email):
            pass

        @abstractmethod
        def save_oauthtokens(self, oauthtokens):
            pass

        @abstractmethod
        def delete_oauthtokens(self, user_email):
            pass
else:
    from abc import ABCMeta, abstractmethod
    class AbstractZohoOAuthPersistence():
        __metaclass__ = ABCMeta

        @abstractmethod
        def get_oauthtokens(self, user_email):
            pass

        @abstractmethod
        def save_oauthtokens(self, oauthtokens):
            pass

        @abstractmethod
        def delete_oauthtokens(self, user_email):
            pass


'''
Created on Aug 11, 2017

@author: sumanth-3058
'''

class PathIdentifier(object):
    '''
    classdocs
    '''


    def __init__(self):
        '''
        Constructor
        '''
        
    @staticmethod
    def get_client_library_root():
        import os
        return os.path.dirname(__file__)
